from datetime import datetime

from pydantic import BaseModel, AnyHttpUrl


# noinspection PyPep8Naming
class _AWS_Identity_Center_Client_Auth(BaseModel):
    """
    Stores the results of the RegisterClient API call.
    This model is to aid possible persistence of this data.
    https://docs.aws.amazon.com/singlesignon/latest/OIDCAPIReference/API_RegisterClient.html
    """

    authorizationEndpoint: AnyHttpUrl
    """
    The endpoint where the client can request authorization.
    """

    clientId: str
    """
    The unique identifier string for each client. This client uses this identifier to get authenticated by the service in subsequent calls.
    """

    clientIdIssuedAt: int
    """
    Indicates the time at which the clientId and clientSecret were issued.
    """

    clientSecret: str
    """
    A secret string generated for the client. The client will use this string to get authenticated by the service in subsequent calls.
    """

    clientSecretExpiresAt: int
    """
    Indicates the time at which the clientId and clientSecret will become invalid.
    """

    tokenEndpoint: AnyHttpUrl
    """
    The endpoint where the client can get an access token.
    """

    @property
    def clientIdIssuedAtDt(self):
        return datetime.fromtimestamp(self.clientSecretExpiresAt)

    @property
    def clientSecretExpiresAtDt(self):
        return datetime.fromtimestamp(self.clientSecretExpiresAt)



from typing import *

from pydantic import PrivateAttr

try:
    import boto3
except ImportError:
    raise ImportError("AWS_Session requires boto3 to be installed")

if TYPE_CHECKING:
    from config_wrangler.config_templates.aws.s3_bucket import S3_Bucket, S3_Bucket_Key

# NOTE: If you are not seeing boto3-stubs code completion in Intellij-based IDEs,
#       please increase the intellisense filesize limit
#       e.g `idea.max.intellisense.filesize=20000` in IDE custom properties
#       (Help > Edit Custom Properties), then restart.

from config_wrangler.config_templates.credentials import Credentials


# AWSSession does not look right, so we added underscores
# noinspection PyPep8Naming
class AWS_Session(Credentials):
    # sso_session: Optional[DynamicallyReferenced] = None
    region_name: Optional[str] = None

    _session: boto3.session.Session = PrivateAttr(default=None)
    _service: str = PrivateAttr(default=None)

    @property
    def session(self) -> boto3.session.Session:
        if self._session is None:
            self._session = boto3.session.Session(
                aws_access_key_id=self.user_id,
                aws_secret_access_key=self.get_password(),
                region_name=self.region_name,
            )
        return self._session

    def set_session(self, session: boto3.session.Session):
        self._session = session

    @property
    def has_session(self) -> bool:
        return self._session is not None

    @property
    def resource(self):
        return self._get_resource()

    @property
    def client(self):
        return self._get_client()

    def _get_resource(self, service: str = None):
        if service is None:
            service = self._service
        # noinspection PyTypeChecker
        return self.session.resource(service, region_name=self.region_name)

    def _get_client(self, service: str = None):
        if service is None:
            service = self._service
        # noinspection PyTypeChecker
        return self.session.client(service, region_name=self.region_name)

    def get_copy(self, copied_by: str = 'get_copy') -> 'AWS_Session':
        return self._factory(cls=self.__class__)

    def _factory(self, cls, exclude: Set[str] = None, **attributes):
        if exclude is None:
            exclude = set()
        exclude.update(attributes.keys())
        new_object = cls.model_construct(
            **attributes,
            **self._dict_for_init(exclude=exclude)
        )
        self.add_child(str(cls), new_object)
        if self.has_session:
            new_object.set_session(self.session)
        return new_object

    def nav_to_bucket(self, bucket_name) -> 'S3_Bucket':
        from config_wrangler.config_templates.aws.s3_bucket import S3_Bucket
        return self._factory(
            S3_Bucket,
            exclude={'bucket_name', 'folder', 'file_name', 'key'},
            bucket_name=bucket_name,
        )

    @staticmethod
    def split_s3_uri(s3_uri: str) -> Tuple[str, str]:
        parts = s3_uri.split("/", 3)
        # Note 'S3://bucket-name/key-name/file.txt'.split('/', 3)
        # Returns ['S3:', '', 'bucket-name', 'key-name/file.txt']
        if parts[0].lower() != 's3:' or parts[1] != '':
            raise ValueError(f"S3 URI '{s3_uri}' does not appear to be valid")
        if len(parts) == 3:
            return parts[2], ''
        else:
            return parts[2], parts[3]

    def nav_to_s3_link(self, s3_uri: str) -> 'S3_Bucket_Key':
        bucket, key = self.split_s3_uri(s3_uri)
        return self.nav_to_bucket(bucket) / key

    def get_ssm(self):
        from config_wrangler.config_templates.aws.ssm import SSM
        return self._factory(
            SSM,
        )

    def get_secrets_manager(self):
        from config_wrangler.config_templates.aws.secrets_manager import SecretsManager
        return self._factory(
            SecretsManager,
        )



from datetime import datetime

from pydantic import BaseModel, AnyHttpUrl, PrivateAttr

from config_wrangler.config_templates.aws._aws_identity_center_client_auth import _AWS_Identity_Center_Client_Auth
from config_wrangler.config_templates.aws.aws_session import AWS_Session


# noinspection PyPep8Naming
class AWS_Identity_Center_Session(AWS_Session):
    ic_client_name: str
    ic_client_type: str = 'public'
    """
    The type of client. The service supports only public as a client type. 
    Anything other than public will be rejected by the service.
    """

    _ic_client_auth: _AWS_Identity_Center_Client_Auth = PrivateAttr(default=None)
    """
    In memory persisted results of the RegisterClient API call
    https://docs.aws.amazon.com/singlesignon/latest/OIDCAPIReference/API_RegisterClient.html
    """



# See https://docs.aws.amazon.com/singlesignon/latest/userguide/howtogetcredentials.html

def get_role_credentials(sso, role_name, account_id):
    role_cred_response = sso.get_role_credentials(
        roleName=role_name,
        accountId=account_id,
        accessToken=access_token,
    )
    if role_cred_response['ResponseMetadata']['HTTPStatusCode'] != 200:
        raise ValueError(f"get_role_credentials response = {role_cred_response['ResponseMetadata']['HTTPStatusCode']}")
    role_creds = role_cred_response['roleCredentials']
    expires_at = datetime.fromtimestamp(role_creds['expiration']/1000)
    print(f"Creds expire at = {expires_at}")


def main():
    from time import sleep
    import webbrowser
    from boto3.session import Session

    ################
    # Config items
    # 1) SSO general
    sso_client_name = 'test_ic_etl'
    sso_start_url = 'https://baosystems.awsapps.com/start'
    region = 'us-east-1'

    # 2) Assume role details
    account_id = '973612564635'
    role_name = 'AdministratorAccess'

    # Static
    sso_client_type = 'public'

    sso_session = Session(region_name=region)
    sso_oidc = sso_session.client('sso-oidc')
    client_creds = sso_oidc.register_client(
        clientName=sso_client_name,
        clientType=sso_client_type,
    )
    device_authorization = sso_oidc.start_device_authorization(
        clientId=client_creds['clientId'],
        clientSecret=client_creds['clientSecret'],
        startUrl=sso_start_url,
    )
    completion_url = device_authorization['verificationUriComplete']
    device_code = device_authorization['deviceCode']
    expires_in = device_authorization['expiresIn']
    interval = device_authorization['interval']
    webbrowser.open(completion_url, autoraise=True)
    token = None

    # Try getting a token until complete or expired
    for n in range(1, expires_in // interval + 1):
        try:
            token = sso_oidc.create_token(
                grantType='urn:ietf:params:oauth:grant-type:device_code',
                deviceCode=device_code,
                clientId=client_creds['clientId'],
                clientSecret=client_creds['clientSecret'],
            )
            # Success, break from the loop
            break
        except sso_oidc.exceptions.AuthorizationPendingException:
            pass
        sleep(interval)

    if token is None:
        raise RuntimeError("Unable to complete authentication in time")

    access_token = token['accessToken']
    sso = sso_session.client('sso')
    account_roles = sso.list_account_roles(
        accessToken=access_token,
        accountId=account_id,
    )
    # {
    #     'nextToken': 'string',
    #     'roleList': [
    #         {
    #             'roleName': 'string',
    #             'accountId': 'string'
    #         },
    #     ]
    # }
    roles = account_roles['roleList']

    #  Find the correct role by name / id
    role = None
    for candidate_role in roles:
        print(f"candidate_role = {candidate_role}")
        if candidate_role['roleName'] == role_name:
            role = candidate_role
    if role is None:
        raise ValueError(f"Session {sso_session} does not have access to role {role_name}")

    role_cred_response = sso.get_role_credentials(
        roleName=role_name,
        accountId=account_id,
        accessToken=access_token,
    )
    if role_cred_response['ResponseMetadata']['HTTPStatusCode'] != 200:
        raise ValueError(f"get_role_credentials response = {role_cred_response['ResponseMetadata']['HTTPStatusCode']}")
    role_creds = role_cred_response['roleCredentials']
    expires_at = datetime.fromtimestamp(role_creds['expiration']/1000)
    print(f"Creds expire at = {expires_at}")

    session = Session(
        region_name=region,
        aws_access_key_id=role_creds['accessKeyId'],
        aws_secret_access_key=role_creds['secretAccessKey'],
        aws_session_token=role_creds['sessionToken'],
    )

    print(session)

# TOTP Example
# https://kevinhakanson.com/2017-10-21-creating-and-using-an-aws-virtual-mfa-device-with-the-aws-sdk-for-python/
def login_totp():
    sts_client = session.client('sts')

    totp = pyotp.TOTP(string_seed)
    token_code = totp.now()

    response = sts_client.assume_role(
        RoleArn='arn:aws:iam::123456789012:role/kjh-DuperRole',
        RoleSessionName='my-python-session',
        SerialNumber='arn:aws:iam::123456789012:mfa/service-user/kjh-SuperDuperUser',
        TokenCode=token_code
    )


if __name__ == '__main__':
    main()
